#include "tracer.h"

#define _GNU_SOURCE
#include <errno.h>
#include <error.h>
#include <fcntl.h>
#include <getopt.h>
#include <limits.h>
#include <linux/unistd.h>
#include <pthread.h>
#include <sched.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <time.h>
#include <unistd.h>

/*
 * From cyclictest code source
 */

#define KVARS 32
#define KVARNAMELEN 32
#define KVALUELEN 32

#define MAX_PATH 255

static char *fileprefix;
static char *procfileprefix = "/proc/sys/kernel/";
static char *debugfileprefix = "/sys/kernel/debug/tracing/";

static int trace_fd = -1;
static int tracemark_fd = -1;

static int kernvar(int mode, const char *name, char *value,
                   size_t sizeofvalue) {
  char filename[128];
  int retval = 1;
  int path;
  size_t len_prefix = strlen(fileprefix), len_name = strlen(name);

  memcpy(filename, fileprefix, len_prefix);
  memcpy(filename + len_prefix, name, len_name + 1);

  path = open(filename, mode);
  if (path >= 0) {
    if (mode == O_RDONLY) {
      int got;
      if ((got = read(path, value, sizeofvalue)) > 0) {
        retval = 0;
        value[got - 1] = '\0';
      }
    } else if (mode == O_WRONLY) {
      if (write(path, value, sizeofvalue) == sizeofvalue) retval = 0;
    }
    close(path);
  }
  return retval;
}

static void setkernvar(const char *name, char *value) {
  if (kernvar(O_WRONLY, name, value, strlen(value)))
    fprintf(stderr, "could not set %s to %s\n", name, value);
}

void open_fds(void) {

  fileprefix = debugfileprefix;

  char trace_path[MAX_PATH];
  char tracemark_path[MAX_PATH];

  // Open tracing_on file
  strcpy(trace_path, fileprefix);
  strcat(trace_path, "tracing_on");
  if ((trace_fd = open(trace_path, O_WRONLY)) == -1)
    printf("unable to open %s for tracing", trace_path);

  // Open trace mark file
  strcpy(tracemark_path, fileprefix);
  strcat(tracemark_path, "trace_marker");
  if ((tracemark_fd = open(tracemark_path, O_WRONLY)) == -1)
    printf("unable to open %s for tracing", tracemark_path);

}

void setup_tracer(int enable_graph) {

  fileprefix = procfileprefix;
  setkernvar("ftrace_enabled", "1");

  fileprefix = debugfileprefix;
  // Clear old traces by setting tracer to nop first
  setkernvar("current_tracer", "nop");
  if(enable_graph)
    setkernvar("current_tracer", "function_graph");
  else
    setkernvar("current_tracer", "function");

  open_fds();

  tracing(0);
}

void tracing(int on) {
  if (on)
    write(trace_fd, "1", 1);
  else
    write(trace_fd, "0", 1);
}

void tracemark(char * s) {
  write(tracemark_fd, s, strlen(s));
}
